""""harness for computing inverse wavelet transform using time transform, take input .dat file in wavelet domain (Nt rows by Nf columns)
and write to .dat file in time domain (columns frequency, h(t))"""
import sys
from time import perf_counter
import numpy as np

from WDMWaveletTransforms.wavelet_transforms import inverse_wavelet_time

def main():
    #assume input .dat file is Nt rows by Nf columns
    if len(sys.argv)!=5:
        print("inverse_wavelet_time_harness.py filename_wavelet_in filename_time_out dt mult")
        sys.exit(1)

    file_in = sys.argv[1]
    file_out = sys.argv[2]
    dt = np.float64(sys.argv[3])
    mult = np.int64(sys.argv[4])

    print('begin loading data file')
    t0 = perf_counter()
    wave_in = np.loadtxt(file_in)
    t1 = perf_counter()
    print('loaded input file in %5.3fs'%(t1-t0))

    Nt = wave_in.shape[0]
    Nf = wave_in.shape[1]

    ND = Nt*Nf
    Tobs = dt*ND

    #time and frequency grids
    ts = np.arange(0, ND)*dt

    t0 = perf_counter()
    signal_time = inverse_wavelet_time(wave_in, Nf, Nt, mult=mult)
    t1 = perf_counter()

    print('got time domain transform in %5.3fs'%(t1-t0))

    t4 = perf_counter()
    np.savetxt(file_out, np.vstack([ts, signal_time]).T)
    t5 = perf_counter()
    print('saved file in %5.3fs'%(t5-t4))

if __name__=='__main__':
    main()


""""harness for computing forward frequency domain wavelet transform, take input .dat file in frequency domain (columns frequency, real part(h(f)), imag part(h(f))"
write to .dat file in wavelet domain (Nt rows by Nf columns)"""

import sys
from time import perf_counter
import numpy as np

from WDMWaveletTransforms.wavelet_transforms import transform_wavelet_freq

def main():
    if len(sys.argv)!=6:
        print("forward_wavelet_freq_harness.py filename_freq_in filename_wavelet_out dt Nt Nf")
        sys.exit(1)

    #transform parameters
    file_in = sys.argv[1]
    file_out = sys.argv[2]

    dt = np.float64(sys.argv[3])
    Nt = np.int64(sys.argv[4])
    Nf = np.int64(sys.argv[5])

    print('begin loading data file')
    t0 = perf_counter()
    #the frequency domain representation
    fs_in,signal_freq_real,signal_freq_im = np.loadtxt(file_in).T
    signal_freq = signal_freq_real+1j*signal_freq_im
    t1 = perf_counter()
    print('loaded input file in %5.3fs'%(t1-t0))

    ND = Nt*Nf
    Tobs = dt*ND

    #time and frequency grids
    fs = np.arange(0,ND//2+1)*1/(Tobs)
    assert np.all(fs==fs_in)

    t0 = perf_counter()
    wave_freq = transform_wavelet_freq(signal_freq,Nf,Nt)
    t1 = perf_counter()

    print('got frequency domain transform in %5.3fs'%(t1-t0))

    t4 = perf_counter()
    np.savetxt(file_out,wave_freq)
    t5 = perf_counter()
    print('saved file in %5.3fs'%(t5-t4))

if __name__=='__main__':
    main()


"""helper to make sure available fft functions are consistent across modules depending on install
mkl-fft is faster so it is the default, but numpy fft is probably more commonly installed to it is the fallback"""
try:
    print('using mkl fft')
    import mkl_fft
    rfft = mkl_fft.rfft_numpy
    irfft = mkl_fft.irfft_numpy
    fft = mkl_fft.fft
    ifft = mkl_fft.ifft
except ImportError:
    print('mkl fft not available trying numpy')
    import numpy
    rfft = numpy.fft.rfft
    irfft = numpy.fft.irfft
    fft = numpy.fft.fft
    ifft = numpy.fft.ifft



"""functions for computing the inverse wavelet transforms"""
from numba import njit
import numpy as np
import WDMWaveletTransforms.fft_funcs as fft


# @njit()
def inverse_wavelet_freq_helper_fast(wave_in, phif, Nf, Nt):
    """jit compatible loop for inverse_wavelet_freq"""
    ND = Nf*Nt

    prefactor2s = np.zeros(Nt, np.complex128)
    res = np.zeros(ND//2+1, dtype=np.complex128)

    for m in range(0, Nf+1):
        pack_wave_inverse(m, Nt, Nf, prefactor2s, wave_in)
        # with numba.objmode(fft_prefactor2s="complex128[:]"):
        fft_prefactor2s = fft.fft(prefactor2s)
        unpack_wave_inverse(m, Nt, Nf, phif, fft_prefactor2s, res)

    return res


@njit()
def unpack_wave_inverse(m, Nt, Nf, phif, fft_prefactor2s, res):
    """helper for unpacking results of frequency domain inverse transform"""

    if m == 0 or m == Nf:
        for i_ind in range(0, Nt//2):
            i = np.abs(m*Nt//2-i_ind)  # i_off+i_min2
            ind3 = (2*i) % Nt
            res[i] += fft_prefactor2s[ind3]*phif[i_ind]
        if m == Nf:
            i_ind = Nt//2
            i = np.abs(m*Nt//2-i_ind)  # i_off+i_min2
            ind3 = 0
            res[i] += fft_prefactor2s[ind3]*phif[i_ind]
    else:
        ind31 = (Nt//2*m) % Nt
        ind32 = (Nt//2*m) % Nt
        for i_ind in range(0, Nt//2):
            i1 = Nt//2*m-i_ind
            i2 = Nt//2*m+i_ind
            # assert ind31 == i1 % Nt
            # assert ind32 == i2 % Nt
            res[i1] += fft_prefactor2s[ind31]*phif[i_ind]
            res[i2] += fft_prefactor2s[ind32]*phif[i_ind]
            ind31 -= 1
            ind32 += 1
            if ind31 < 0:
                ind31 = Nt-1
            if ind32 == Nt:
                ind32 = 0

        res[Nt//2*m] = fft_prefactor2s[(Nt//2*m) % Nt]*phif[0]

# @njit()
# def unpack_wave_inverse(m, Nt, Nf, phif, fft_prefactor2s, res):
#    """helper for unpacking results of frequency domain inverse transform"""
#    ND = Nt*Nf
#    i_min2 = min(max(Nt//2*(m-1), 0), ND//2+1)
#    i_max2 = min(max(Nt//2*(m+1), 0), ND//2+1)
#    for i in range(i_min2, i_max2):
#        i_ind = np.abs(i-Nt//2*m)
#        if i_ind > Nt//2:
#            continue
#        if m == 0:
#            res[i] += fft_prefactor2s[(2*i) % Nt]*phif[i_ind]
#        elif m == Nf:
#            res[i] += fft_prefactor2s[(2*i) % Nt]*phif[i_ind]
#        else:
#            res[i] += fft_prefactor2s[i % Nt]*phif[i_ind]


@njit()
def pack_wave_inverse(m, Nt, Nf, prefactor2s, wave_in):
    """helper for fast frequency domain inverse transform to prepare for fourier transform"""
    if m == 0:
        for n in range(0, Nt):
            prefactor2s[n] = 1/np.sqrt(2)*wave_in[(2*n) % Nt, 0]
    elif m == Nf:
        for n in range(0, Nt):
            prefactor2s[n] = 1/np.sqrt(2)*wave_in[(2*n) % Nt+1, 0]
    else:
        for n in range(0, Nt):
            val = wave_in[n, m]
            if (n+m) % 2:
                mult2 = -1j
            else:
                mult2 = 1

            prefactor2s[n] = mult2*val



""""harness for computing inverse wavelet transform using frequency transform+fft, take input .dat file in wavelet domain (Nt rows by Nf columns)
and write to .dat file in time domain (columns frequency, h(t))"""
import sys
from time import perf_counter
import numpy as np

from WDMWaveletTransforms.wavelet_transforms import inverse_wavelet_freq_time

def main():
    #assume input .dat file is Nt rows by Nf columns
    if len(sys.argv)!=4:
        print("inverse_wavelet_freq_time_harness.py filename_wavelet_in filename_time_out dt")
        sys.exit(1)

    file_in = sys.argv[1]
    file_out = sys.argv[2]
    dt = np.float64(sys.argv[3])

    print('begin loading data file')
    t0 = perf_counter()
    wave_in = np.loadtxt(file_in)
    t1 = perf_counter()
    print('loaded input file in %5.3fs'%(t1-t0))

    Nt = wave_in.shape[0]
    Nf = wave_in.shape[1]

    ND = Nt*Nf
    Tobs = dt*ND

    #time and frequency grids
    ts = np.arange(0,ND)*dt

    t0 = perf_counter()
    signal_time = inverse_wavelet_freq_time(wave_in,Nf,Nt)
    t1 = perf_counter()

    print('got time domain transform via freq in %5.3fs'%(t1-t0))

    t4 = perf_counter()
    np.savetxt(file_out,np.vstack([ts,signal_time]).T)
    t5 = perf_counter()
    print('saved file in %5.3fs'%(t5-t4))

if __name__=='__main__':
    main()


""""harness for computing inverse frequency domain wavelet transform, take input .dat file in wavelet domain (Nt rows by Nf columns)
and write to .dat file in frequency domain (columns frequency, real part(h(f)), imag part(h(f))"""
import sys
from time import perf_counter
import numpy as np

from WDMWaveletTransforms.wavelet_transforms import inverse_wavelet_freq

def main():
    #assume input .dat file is Nt rows by Nf columns
    if len(sys.argv)!=4:
        print("inverse_wavelet_freq_harness.py filename_wavelet_in filename_freq_out dt")
        sys.exit(1)

    #transform parameters
    file_in = sys.argv[1]
    file_out = sys.argv[2]

    dt = np.float64(sys.argv[3])

    #get a wavelet representation of a signal
    print('begin loading data file')
    t0 = perf_counter()
    wave_in = np.loadtxt(file_in)
    t1 = perf_counter()
    print('loaded input file in %5.3fs'%(t1-t0))

    Nt = wave_in.shape[0]
    Nf = wave_in.shape[1]

    ND = Nt*Nf
    Tobs = dt*ND

    #time and frequency grids
    ts = np.arange(0,ND)*dt
    fs = np.arange(0,ND//2+1)*1/(Tobs)


    t0 = perf_counter()
    signal_freq = inverse_wavelet_freq(wave_in,Nf,Nt)
    t1 = perf_counter()

    print('got frequency domain transform in %5.3fs'%(t1-t0))

    t4 = perf_counter()
    np.savetxt(file_out,np.vstack([fs,np.real(signal_freq),np.imag(signal_freq)]).T)
    t5 = perf_counter()
    print('saved file in %5.3fs'%(t5-t4))


if __name__=='__main__':
    main()



"""helper functions for transform_time.py"""
import numpy as np
import WDMWaveletTransforms.fft_funcs as fft
from WDMWaveletTransforms.transform_freq_funcs import phitilde_vec_norm, transform_wavelet_freq_helper
from WDMWaveletTransforms.transform_time_funcs import phi_vec, transform_wavelet_time_helper
from WDMWaveletTransforms.inverse_wavelet_freq_funcs import inverse_wavelet_freq_helper_fast
from WDMWaveletTransforms.inverse_wavelet_time_funcs import inverse_wavelet_time_helper_fast


def inverse_wavelet_time(wave_in, Nf, Nt, nx=4., mult=32):
    """fast inverse wavelet transform to time domain"""
    mult = min(mult, Nt//2)  # make sure K isn't bigger than ND
    phi = phi_vec(Nf, nx=nx, mult=mult)/2

    return inverse_wavelet_time_helper_fast(wave_in, phi, Nf, Nt, mult)


def inverse_wavelet_freq_time(wave_in, Nf, Nt, nx=4.):
    """inverse wavlet transform to time domain via fourier transform of frequency domain"""
    res_f = inverse_wavelet_freq(wave_in, Nf, Nt, nx)
    return fft.irfft(res_f)


def inverse_wavelet_freq(wave_in, Nf, Nt, nx=4.):
    """inverse wavelet transform to freq domain signal"""
    phif = phitilde_vec_norm(Nf, Nt, nx)
    return inverse_wavelet_freq_helper_fast(wave_in, phif, Nf, Nt)


def transform_wavelet_time(data, Nf, Nt, nx=4., mult=32):
    """do the wavelet transform in the time domain,
    note there can be significant leakage if mult is too small and the
    transform is only approximately exact if mult=Nt/2"""
    mult = min(mult, Nt//2)  # make sure K isn't bigger than ND
    phi = phi_vec(Nf, nx, mult)
    wave = transform_wavelet_time_helper(data, Nf, Nt, phi, mult)

    return wave


def transform_wavelet_freq_time(data, Nf, Nt, nx=4.):
    """transform time domain data into wavelet domain via fft and then frequency transform"""
    data_fft = fft.rfft(data)

    return transform_wavelet_freq(data_fft, Nf, Nt, nx)


def transform_wavelet_freq(data, Nf, Nt, nx=4.):
    """do the wavelet transform using the fast wavelet domain transform"""
    phif = 2/Nf*phitilde_vec_norm(Nf, Nt, nx)
    return transform_wavelet_freq_helper(data, Nf, Nt, phif)



"""helper functions for transform_time.py"""
import numpy as np
from numba import njit
from WDMWaveletTransforms.transform_freq_funcs import phitilde_vec
import WDMWaveletTransforms.fft_funcs as fft


def transform_wavelet_time_helper(data, Nf, Nt, phi, mult):
    """helper function do do the wavelet transform in the time domain"""
    # the time domain data stream
    ND = Nf*Nt

    # mult, can cause bad leakage if it is too small but may be possible to mitigate
    # Filter is mult times pixel with in time

    K = mult*2*Nf

    # windowed data packets
    wdata = np.zeros(K)

    wave = np.zeros((Nt, Nf))  # wavelet wavepacket transform of the signal
    data_pad = np.zeros(ND+K)
    data_pad[:ND] = data
    data_pad[ND:ND+K] = data[:K]

    for i in range(0, Nt):
        assign_wdata(i, K, ND, Nf, wdata, data_pad, phi)
        wdata_trans = fft.rfft(wdata, K)
        pack_wave(i, mult, Nf, wdata_trans, wave)

    return wave


@njit()
def assign_wdata(i, K, ND, Nf, wdata, data_pad, phi):
    """assign wdata to be fftd in loop, data_pad needs K extra values on the right to loop"""
    # half_K = np.int64(K/2)
    jj = i*Nf-K//2
    if jj < 0:
        jj += ND  # periodically wrap the data
    if jj >= ND:
        jj -= ND  # periodically wrap the data
    for j in range(0, K):
        # jj = i*Nf-half_K+j
        wdata[j] = data_pad[jj]*phi[j]  # apply the window
        jj += 1
        # if jj == ND:
        #    jj -= ND # periodically wrap the data


@njit()
def pack_wave(i, mult, Nf, wdata_trans, wave):
    """pack fftd wdata into wave array"""
    if i % 2 == 0 and i < wave.shape[0]-1:
        # m=0 value at even Nt and
        wave[i, 0] = np.real(wdata_trans[0])/np.sqrt(2)
        wave[i+1, 0] = np.real(wdata_trans[Nf*mult])/np.sqrt(2)

    for j in range(1, Nf):
        if (i+j) % 2:
            wave[i, j] = -np.imag(wdata_trans[j*mult])
        else:
            wave[i, j] = np.real(wdata_trans[j*mult])


def phi_vec(Nf, nx=4., mult=16):
    """get time domain phi as fourier transform of phitilde_vec"""
    # TODO fix mult

    OM = np.pi
    DOM = OM/Nf
    insDOM = 1./np.sqrt(DOM)
    K = mult*2*Nf
    half_K = mult*Nf  # np.int64(K/2)

    dom = 2*np.pi/K  # max frequency is K/2*dom = pi/dt = OM

    DX = np.zeros(K, dtype=np.complex128)

    # zero frequency
    DX[0] = insDOM

    DX = DX.copy()
    # postive frequencies
    DX[1:half_K+1] = phitilde_vec(dom*np.arange(1, half_K+1), Nf, nx)
    # negative frequencies
    DX[half_K+1:] = phitilde_vec(-dom*np.arange(half_K-1, 0, -1), Nf, nx)
    DX = K*fft.ifft(DX, K)

    phi = np.zeros(K)
    phi[0:half_K] = np.real(DX[half_K:K])
    phi[half_K:] = np.real(DX[0:half_K])

    nrm = np.sqrt(K/dom)  # *np.linalg.norm(phi)

    fac = np.sqrt(2.0)/nrm
    phi *= fac
    return phi



"""helper functions for transform_freq"""
import numpy as np
from numba import njit
import scipy.special
import WDMWaveletTransforms.fft_funcs as fft


def phitilde_vec(om, Nf, nx=4.):
    """compute phitilde, om i array, nx is filter steepness, defaults to 4."""
    OM = np.pi   # Nyquist angular frequency
    DOM = OM/Nf  # 2 pi times DF
    insDOM = 1./np.sqrt(DOM)
    B = OM/(2*Nf)
    A = (DOM-B)/2
    z = np.zeros(om.size)

    mask = (np.abs(om) >= A) & (np.abs(om) < A+B)

    x = (np.abs(om[mask])-A)/B
    y = scipy.special.betainc(nx, nx, x)
    z[mask] = insDOM*np.cos(np.pi/2.*y)

    z[np.abs(om) < A] = insDOM
    return z


def phitilde_vec_norm(Nf, Nt, nx):
    """normalize phitilde as needed for inverse frequency domain transform"""
    ND = Nf*Nt
    oms = 2*np.pi/ND*np.arange(0, Nt//2+1)
    phif = phitilde_vec(oms, Nf, nx)
    # nrm should be 1
    nrm = np.sqrt((2*np.sum(phif[1:]**2)+phif[0]**2)*2*np.pi/ND)
    nrm /= np.pi**(3/2)/np.pi
    phif /= nrm
    return phif


@njit()
def tukey(data, alpha, N):
    """apply tukey window function to data"""
    imin = np.int64(alpha*(N-1)/2)
    imax = np.int64((N-1)*(1-alpha/2))
    Nwin = N-imax

    for i in range(0, N):
        f_mult = 1.0
        if i < imin:
            f_mult = 0.5*(1.+np.cos(np.pi*(i/imin-1.)))
        if i > imax:
            f_mult = 0.5*(1.+np.cos(np.pi/Nwin*(i-imax)))
        data[i] *= f_mult


def transform_wavelet_freq_helper(data, Nf, Nt, phif):
    """helper to do the wavelet transform using the fast wavelet domain transform"""
    wave = np.zeros((Nt, Nf))  # wavelet wavepacket transform of the signal

    DX = np.zeros(Nt, dtype=np.complex128)
    for m in range(0, Nf+1):
        DX_assign_loop(m, Nt, Nf, DX, data, phif)
        DX_trans = fft.ifft(DX, Nt)
        DX_unpack_loop(m, Nt, Nf, DX_trans, wave)
    return wave


@njit()
def DX_assign_loop(m, Nt, Nf, DX, data, phif):
    """helper for assigning DX in the main loop"""
    i_base = Nt//2
    jj_base = m*Nt//2

    if m == 0 or m == Nf:
        # NOTE this term appears to be needed to recover correct constant (at least for m=0) but was previously missing
        DX[Nt//2] = phif[0]*data[m*Nt//2]/2.
        DX[Nt//2] = phif[0]*data[m*Nt//2]/2.
    else:
        DX[Nt//2] = phif[0]*data[m*Nt//2]
        DX[Nt//2] = phif[0]*data[m*Nt//2]

    for jj in range(jj_base+1-Nt//2, jj_base+Nt//2):
        j = np.abs(jj-jj_base)
        i = i_base-jj_base+jj
        if m == Nf and jj > jj_base:
            DX[i] = 0.
        elif m == 0 and jj < jj_base:
            DX[i] = 0.
        elif j == 0:
            continue
        else:
            DX[i] = phif[j]*data[jj]


@njit()
def DX_unpack_loop(m, Nt, Nf, DX_trans, wave):
    """helper for unpacking fftd DX in main loop"""
    if m == 0:
        # half of lowest and highest frequency bin pixels are redundant
        # so store them in even and odd components of m=0 respectively
        for n in range(0, Nt, 2):
            wave[n, 0] = np.real(DX_trans[n]*np.sqrt(2))
    elif m == Nf:
        for n in range(0, Nt, 2):
            wave[n+1, 0] = np.real(DX_trans[n]*np.sqrt(2))
    else:
        for n in range(0, Nt):
            if m % 2:
                if (n+m) % 2:
                    wave[n, m] = -np.imag(DX_trans[n])
                else:
                    wave[n, m] = np.real(DX_trans[n])
            else:
                if (n+m) % 2:
                    wave[n, m] = np.imag(DX_trans[n])
                else:
                    wave[n, m] = np.real(DX_trans[n])



""""harness for computing forward time domain wavelet transform via fft, take input .dat file in time domain (columns ftime, h(t)"
write to .dat file in wavelet domain (Nt rows by Nf columns)"""

import sys
from time import perf_counter
import numpy as np

from WDMWaveletTransforms.wavelet_transforms import transform_wavelet_freq_time

def main():
    if len(sys.argv)!=6:
        print("forward_wavelet_freq_time_harness.py filename_time_in filename_wavelet_out dt Nt Nf")
        sys.exit(1)

    #transform parameters
    file_in = sys.argv[1]
    file_out = sys.argv[2]

    dt = np.float64(sys.argv[3])
    Nt = np.int64(sys.argv[4])
    Nf = np.int64(sys.argv[5])
    mult = np.int64(sys.argv[5])

    print('begin loading data file')
    t0 = perf_counter()
    #the frequency domain representation
    ts_in,signal_time = np.loadtxt(file_in).T
    t1 = perf_counter()
    print('loaded input file in %5.3fs'%(t1-t0))

    ND = Nt*Nf
    Tobs = dt*ND

    #time and frequency grids
    ts = dt*np.arange(0,ND)
    assert np.all(ts==ts_in)

    t0 = perf_counter()
    wave_time = transform_wavelet_freq_time(signal_time,Nf,Nt)
    t1 = perf_counter()

    print('got time domain transform in %5.3fs'%(t1-t0))

    t4 = perf_counter()
    np.savetxt(file_out,wave_time)
    t5 = perf_counter()
    print('saved file in %5.3fs'%(t5-t4))


if __name__=='__main__':
    main()


"""functions for computing the inverse wavelet transforms"""
from numba import njit
import numpy as np

import WDMWaveletTransforms.fft_funcs as fft


def inverse_wavelet_time_helper_fast(wave_in, phi, Nf, Nt, mult):
    """helper loop for fast inverse wavelet transform"""
    ND = Nf*Nt
    K = mult*2*Nf
    # res = np.zeros(ND)

    # extend this array, we can use wrapping boundary conditions at end
    res = np.zeros(ND+K+Nf)

    afins = np.zeros(2*Nf, dtype=np.complex128)

    for n in range(0, Nt):
        # old unpacked way, should still work but is necessarily slower,
        # might be more comparable if it could be written as an irfft instead
        # pack_wave_time_helper(n, Nf, Nt, wave_in, afins)
        # ffts_fin_real = np.real(fft.fft(afins))
        # unpack_time_wave_helper(n, Nf, Nt, K, phi, ffts_fin_real, res)

        # we can pack both the sin and cos parts into the real and imaginary parts
        # of the same transform so we only need to do every other one
        # this currently assumes Nt is even
        if n % 2 == 0:
            pack_wave_time_helper_compact(n, Nf, Nt, wave_in, afins)
            ffts_fin = fft.fft(afins)
            unpack_time_wave_helper_compact(n, Nf, Nt, K, phi, ffts_fin, res)

    # wrap boundary conditions
    res[:min(K+Nf, ND)] += res[ND:min(ND+K+Nf, 2*ND)]
    if K+Nf > ND:
        res[:K+Nf-ND] += res[2*ND:ND+K*Nf]

    res = res[:ND]

    return res


@njit()
def unpack_time_wave_helper(n, Nf, Nt, K, phis, fft_fin_real, res):
    """helper for time domain wavelet transform to unpack wavelet domain coefficients"""
    ND = Nf*Nt

    idxf = (-K//2+n*Nf+ND) % (2*Nf)
    k = (-K//2+n*Nf) % ND

    for k_ind in range(0, K):
        res_loc = fft_fin_real[idxf]
        res[k] += phis[k_ind]*res_loc
        idxf += 1
        k += 1

        if idxf == 2*Nf:
            idxf = 0
        if k == ND:
            k = 0


@njit()
def unpack_time_wave_helper_compact(n, Nf, Nt, K, phis, fft_fin, res):
    """helper for time domain wavelet transform to unpack wavelet domain coefficients
    in compact representation where cosine and sine parts are real and imaginary parts"""
    ND = Nf*Nt
    fft_fin_real = np.zeros(4*Nf)
    fft_fin_imag = np.zeros(4*Nf)
    for itrf in range(0, 2*Nf):
        fft_fin_real[itrf] = np.real(fft_fin[itrf])
        fft_fin_real[itrf+2*Nf] = fft_fin_real[itrf]
        fft_fin_imag[itrf] = np.imag(fft_fin[(itrf+Nf) % (2*Nf)])
        fft_fin_imag[itrf+2*Nf] = fft_fin_imag[itrf]

    idxf1_base = (-K//2+n*Nf+ND) % (2*Nf)
    k1_base = (-K//2+n*Nf) % ND
    for k_ind in range(0, K, 2*Nf):
        for idxf1_add in range(0, 2*Nf):
            idxf1 = idxf1_base+idxf1_add  # k_ind % (2*Nf)
            k_ind_loc = k_ind+idxf1_add
            k1 = k1_base+k_ind_loc

            res[k1] += phis[k_ind_loc]*fft_fin_real[idxf1]
            res[k1+Nf] += phis[k_ind_loc]*fft_fin_imag[idxf1]

# @njit()
# def pack_wave_time_helper(n, Nf, Nt, wave_in, afins):
#    """helper for time domain transform to pack wavelet domain coefficients"""
#    if n % 2 == 0:
#        # assign highest and lowest bin correctly
#        afins[0] = 1/np.sqrt(2)*wave_in[n, 0]
#        if n+1 < Nt:
#            afins[Nf] = 1/np.sqrt(2)*wave_in[n+1, 0]
#    else:
#        afins[0] = 0.
#        afins[Nf] = 0.
#
#    for idxm in range(0, Nf//2-1):
#        if n % 2:
#            afins[2*idxm+2] = 1j*wave_in[n, 2*idxm+2]
#        else:
#            afins[2*idxm+2] = wave_in[n, 2*idxm+2]
#
#    for idxm in range(0, Nf//2):
#        if n % 2:
#            afins[2*idxm+1] = -wave_in[n, 2*idxm+1]
#        else:
#            afins[2*idxm+1] = 1j*wave_in[n, 2*idxm+1]


@njit()
def pack_wave_time_helper(n, Nf, Nt, wave_in, afins):
    """helper for time domain transform to pack wavelet domain coefficients"""
    if n % 2 == 0:
        # assign highest and lowest bin correctly
        afins[0] = np.sqrt(2)*wave_in[n, 0]
        if n+1 < Nt:
            afins[Nf] = np.sqrt(2)*wave_in[n+1, 0]
    else:
        afins[0] = 0.
        afins[Nf] = 0.

    for idxm in range(0, Nf//2-1):
        if n % 2:
            afins[2*idxm+2] = 1j*wave_in[n, 2*idxm+2]
            afins[2*Nf-2*idxm-2] = -1j*wave_in[n, 2*idxm+2]
        else:
            afins[2*idxm+2] = 1*wave_in[n, 2*idxm+2]
            afins[2*Nf-2*idxm-2] = 1*wave_in[n, 2*idxm+2]

    for idxm in range(0, Nf//2):
        if n % 2:
            afins[2*idxm+1] = -1*wave_in[n, 2*idxm+1]
            afins[2*Nf-2*idxm-1] = -1*wave_in[n, 2*idxm+1]
        else:
            afins[2*idxm+1] = 1j*wave_in[n, 2*idxm+1]
            afins[2*Nf-2*idxm-1] = -1j*wave_in[n, 2*idxm+1]


@njit()
def pack_wave_time_helper_compact(n, Nf, Nt, wave_in, afins):
    """helper for time domain transform to pack wavelet domain coefficients
    in packed representation with odd and even coefficients in real and imaginary pars"""
    afins[0] = np.sqrt(2)*wave_in[n, 0]
    if n+1 < Nt:
        afins[Nf] = np.sqrt(2)*wave_in[n+1, 0]

    for idxm in range(0, Nf-2, 2):
        afins[idxm+2] = wave_in[n, idxm+2]-wave_in[n+1, idxm+2]
        afins[2*Nf-idxm-2] = wave_in[n, idxm+2]+wave_in[n+1, idxm+2]

        afins[idxm+1] = 1j*(wave_in[n, idxm+1]-wave_in[n+1, idxm+1])
        afins[2*Nf-idxm-1] = -1j*(wave_in[n, idxm+1]+wave_in[n+1, idxm+1])

    afins[Nf-1] = 1j*(wave_in[n, Nf-1]-wave_in[n+1, Nf-1])
    afins[Nf+1] = -1j*(wave_in[n, Nf-1]+wave_in[n+1, Nf-1])
